%% ESPIRiT Reconstruction Demo
% This is a demo on how to generate ESPIRiT maps and use them to perform
% ESPIRiT reconstruction for parallel imaging. It is based on the paper
% Uecker et. al, MRM 2013 DOI 10.1002/mrm.24751. ESPIRiT is a method that
% finds the subspace of multi-coil data from a calibration region in
% k-space using a series of eigen-value decompositions in k-space and image
% space. Here we also use the "soft" sense idea (Uecker et. al, "ESPIRiT 
% Reconstruction using Soft-SENSE", Proceedings of the ISMRM 2013, pp-127)
% by using the eigen values to weight the eigen-vectors. 

clc
clear all
close all


%% Prepare DATA
% Here we perform ESPIRiT calibration on data which has strong aliasing in
% the phase-encode direction. SENSE often fails with this type of data.

load brain_8ch.mat
DATA = DATA/max(max(max(abs(ifft2c(DATA))))) + eps;

ncalib = 24; % use 24 calibration lines to compute compression
ksize = [6,6]; % ESPIRiT kernel-window-size
eigThresh_k = 0.02 % threshold of eigenvectors in k-space
eigThresh_im = 0.9; % threshold of eigenvectors in image space
[sx,sy,Nc] = size(DATA);

% create a sampling mask to simulate x2 undersampling with autocalibration
% lines
mask = zpad(ones(sx,ncalib),[sx,sy]);
mask = repmat(mask,[1,1,8]);
mask(1:2:end,1:3:end,:) = 1;


DATAc = DATA.*mask;
calib = crop(DATAc,[sx,ncalib,Nc]);

%% Compute Eigen-Value Maps
% Maps are computed in two steps. 


% compute Calibration matrix, perform 1st SVD and convert singular vectors
% into k-space kernels

[k,S] = dat2Kernel(calib,ksize);

idx = max(find(S >= S(1)*eigThresh_k));

%%
% crop kernels and compute eigen-value decomposition in image space to get
% maps


[M,W] = kernelEig(k(:,:,:,1:idx),[sx,sy]);


%% Compute Soft-SENSE ESPIRiT Maps 
% crop sensitivity maps according to eigenvalues==1. Note that we have to
% use 2 sets of maps. Here we weight the 2 maps with the eigen-values

maps = M(:,:,:,end-1:end);

% Weight the eigenvectors with soft-senses eigen-values
weights = W(:,:,end-1:end) ;
weights = (weights - eigThresh_im)./(1-eigThresh_im).* (W(:,:,end-1:end) > eigThresh_im);
weights = -cos(pi*weights)/2 + 1/2;

% create and ESPIRiT operator
ESP = ESPIRiT(maps,weights);
nIterCG = 12; 

%% Reconsturctions
% ESPIRiT CG reconstruction with soft-sense and 2 sets of maps
disp('Performing ESPIRiT reconstruction from 2 maps')
tic; [reskESPIRiT, resESPIRiT] = cgESPIRiT(DATAc,ESP, nIterCG, 0.01,DATAc*0); toc

m = sos(resESPIRiT);
cut_factor = 0.99;
cut_index = int32(sx * sy * cut_factor);
sorted = sort(m(:));
cut_value = sorted(cut_index);
m(m > cut_value) = cut_value;

m = m / max(max(m)) * 255;
figure, imshow(m, [0,255])

%Image from complete data
Y=reshape(DATA,sx,sy*Nc);
%[Nneu,MMneu]=size(Y);
%Mneu=MMneu/Nc;
mrec0=zeros(sx,sy);         % initialize mrec0
for j=1:Nc
    Yz=sqrt(sx*sy)*ifftshift(ifft2(ifftshift(Y(:,1+(j-1)*sy:j*sy))));
    mrec0=mrec0+conj(Yz).*Yz;   
end
mrec0=sqrt(mrec0);      %mrec0 is reconstructed image from complete data with SOS

 figure('name','reconstructed image with ESPIRIT'),
 imagesc(m); axis off;
 set(gca,'FontSize',20);
 colormap(gray);
 set(gca,'LooseInset',get(gca,'TightInset'));
 saveas(gcf,'reconstruction_ESPIRIT_3c2r.png')
%saveas(gcf, "ESPIRiT2c2r.png")

 % figure of reconstructed image from complete k-space data (with SoS)
 K=floor(sx*sy/100);                   % for proper scaling of grey values in the figure
 mrec00=reshape(mrec0,sx*sy,1);        % set the largest 1 percent of values to the maximum
 mrec0s=sort(mrec00);                % of the other values
 index=find(mrec00>mrec0s(sx*sy-K));
 mrec00(index)=mrec0s(sx*sy-K);
 mrec00=reshape(mrec00,sx,sy);

figure('name','image reconstruction from full data with SOS'),
imagesc(mrec00); axis off;
set(gca,'FontSize',20);
colormap(gray);

% Comparison with reconstruction from complete data
 m1=mrec00./max(mrec00(:));    % normalization on [0,1]
 mrec=abs(m);
 mrec=mrec./max(mrec(:));    % normalization on [0,1]
% p = psnr1(m1,mrec)          % PSNR value of the error
 
 %Error map
error=abs(mrec-m1)/max(m1(:));  %relative error
 figure('name','error map'), imagesc(error); axis off;
 colormap default
 colorbar
 set(gca,'FontSize',20);
  saveas(gcf,'reconstruction_error_ESPIRIT_3c2r.png')
