function [mrec,crec,Yrec] = mocca_fast_mri(YP,P,P1,M1,L,Nc,input_proj)
% reconstruction from incomplete MRI data (Algorithm 4.1)
%    using the direct solution method from Section 5.1
%
%  Input: YP  acquired coil image data YP=(Y^1.*P ...Y^Nc.*P ) of size N times M*Nc
%         P   projection matrix including the calibration area
%         P1  projection matrix, contains ones according to the structure fixed by
%         L= 2l+1 (odd), the square [-l,l]^2 of size L^2 is the support of
%         trigonometric polynomials for coil sensitivities
%         M1 = calibration matrix dimension in calibration area providing a
%         M1+L-1 box (M1 even)
%         Nc = number of coils
%         input_proj =1,2,3 or 4 depending on the projection matrix (acquired data)
%         for input_proj =1 we need N,M being multiples of 2, for input_p =2 we need N,M
%         being multiples of 4, input_p =3 we need N to be a multiple of 2 and  M being
%         multiple of 8, for input_p =4 we need N being a multiple of 4 and M being
%         multiples of 6
% Output: reconstructed image mrec, coil coefficients crec

%step 1: Build block matrices
[N,MM]=size(YP);
M=MM/Nc;
%fix the calibration data
Y1=[]; %extract the Nc centered calibration matrices of size M1+L-1 x M1+L-1
K1=floor(N/2)-floor((M1+L-1)/2); K2=floor(M/2)-floor((M1+L-1)/2);
for j=1:Nc
    Y1=[Y1 YP(K1+1:K1+M1+L-1,(j-1)*M+K2+1:(j-1)*M+K2+M1+L-1)];
end

%build partial block matrix from shifts of calibration data
l=floor((L-1)/2);
A1=[];
for j=1:Nc
Yb=Y1(:,1+(j-1)*(L+M1-1):j*(L+M1-1));
    for l1=-l:l
        for l2=-l:l
            Ybc=circshift(Yb,[l2,l1]);
            Ybcpart=Ybc(l+1:M1+l,l+1:M1+l); %partial calibration matrix blocks
            Ybv=reshape(Ybcpart,M1^2,1);   %calibration matrix blocks of size M1^2 x L^2
            A1=[A1 Ybv];
        end
    end
end

%Build the Mocca matrix of size M1^2 * Nc x L^2 * Nc
AMocca=[];
ysum=zeros(M1^2,L^2);
for j=1:Nc
    ysum=ysum+A1(:,1+(j-1)*L^2:j*L^2);
end
for k=1:Nc
    Z=[];
    for j=1:Nc
        if j==k
            Z=[Z A1(:,1+(k-1)*L^2:k*L^2)-ysum];
        else
        Z=[Z A1(:,1+(k-1)*L^2:k*L^2)];
        end
    end
    AMocca = [AMocca;Z];
end

% Compute svd of Mocca-Matrix to find the eigenvector c corresponding to
% smallest singular value
[U,S,V]=svd(AMocca);
cvec=V(:,end);                        %vector of sensitivity coefficients

% extract the sensitivity coefficients and reconstruct the sensitivity
% matrices
crec=[]; srec=[]; crec0=[];
for j=1:Nc
    cneu=cvec(1+(j-1)*L^2:j*L^2,1); %partion into coil sensitivity vectors
    cneu=reshape(cneu,L,L);
    crec=[crec cneu];

    % zero padding to obtain an NxM matrix
    K1=floor(N/2)-l; K2=floor(M/2)-l;
    zb=zeros(K1,L); zc=zeros(N+1,K2);
    cneu=[zb; cneu ; zb];
    cneu=[zc cneu zc];
    crec0=[crec0 cneu(1:N,1:M)];
    
    %reconstruct matrix of the sensitivity matrices sj
    fc= sqrt(N*M)*ifftshift(ifft2(ifftshift(crec0(1:N,1+(j-1)*M:j*M))));
    srec=[srec sqrt(M*N)*fc];
end

% Compute vector d in step 4 of Algorithm 3.1 (step 1 of Algorithm 4.1)
d=zeros(N,M);
for j=1:Nc
    d=d+conj(srec(:,1+(j-1)*M:j*M)).*srec(:,1+(j-1)*M:j*M);
end

% compute d^+ (componentwise inverse of d if possible)
if min(d(:))> 10^(-12)
   dinv=ones(N,M)./d;
else
d=reshape(d,N*M,1);
 index=find(abs(d)<10^(-12));  % pay attention if d =0 for some components
 indexg=find(abs(d)>=10^(-12));
 dinv=ones(N*M,1); d0=zeros(N*M,1);
 dinv(indexg)=dinv(indexg)./d(indexg);
 dinv(index)=d0(index);
 dinv=reshape(dinv,N,M);
end


%step 2 of Algorithm 4.1: normalize sensitivities
d12=sqrt(dinv);

for j=1:Nc
    srec(:,1+(j-1)*M:j*M)= d12.*srec(:,1+(j-1)*M:j*M);
end


%Compute the equation system in step 3 of Algorithm 4.1 directly using the
%special structure of the projection matrix P1 (see Section 5.1)
%righthand side
Rs=zeros(N,M);
for j=1:Nc
Rs=Rs+conj(srec(:,1+(j-1)*M:j*M)).*(ifftshift(ifft2(ifftshift(P1.*YP(:,1+(j-1)*M:j*M)))))*sqrt(N*M);
end
% Define G matrix (which decomposes into smaller matrices according to the structure of
% P1 and solve the equation system to compute m
                   
if input_proj==1  %each second column acquired
% Define G matrix (which decomposes into N*M/2 (2x2)-matrices) and solve the system to compute m
m0=zeros(N,M);
for k1=1:N
    for k2=1:M/2
        % define 2x2 partial matrices of G
        G=zeros(2,2); re=zeros(2,1); %initialize G matrix and right-hand side
        for j=1:Nc
            s=[srec(k1,k2+(j-1)*M) srec(k1,k2+M/2+(j-1)*M)];
            G=G+s'*s;
        end
        re=2*[Rs(k1,k2); Rs(k1,k2+M/2)]; %right hand side
        %solve 2x2 system
        mm=linsolve(G,re);
        m0(k1,k2)=mm(1);
        m0(k1,k2+M/2)=mm(2);
    end
end
elseif input_proj==2  % each second row and each second column acquired
                   % define G matrix (which decomposes into N*M/4 (4x4)-matrices) 
                   % and solve the system to compute m
m0=zeros(N,M);
for k1=1:N/2
    for k2=1:M/2
        % define 4x4 partial matrices of G
        G=zeros(4,4); re=zeros(4,1); %initialize G matrix and right-hand side
        for j=1:Nc
            s=[srec(k1+N/2,k2+(j-1)*M) srec(k1+N/2,k2+M/2+(j-1)*M) srec(k1,k2+(j-1)*M)  srec(k1,k2+M/2+(j-1)*M)];
            G=G+s'*s;
        end
        re=4*[Rs(k1+N/2,k2); Rs(k1+N/2,k2+M/2); Rs(k1,k2);  Rs(k1,k2+M/2)]; %right hand side
        % solve 4x4 system
        %mm=linsolve(G,re);
        mm=pinv(G)*re;
        m0(k1+N/2,k2)=mm(1);
        m0(k1+N/2,k2+M/2)=mm(2);
        m0(k1,k2)=mm(3);
        m0(k1,k2+M/2)=mm(4);
    end
end
elseif input_proj==3  % each fourth column acquired
                   % define G matrix (which decomposes into N*M/4 (4x4)-matrices) 
                   % and solve the system to compute m
m0=zeros(N,M);
for k1=1:N
    for k2=1:M/4
        % define 4x4 partial matrices of G
        G=zeros(4,4); re=zeros(4,1); %initialize G matrix and right-hand side
        for j=1:Nc
            s=[srec(k1,k2+(j-1)*M) srec(k1,k2+M/4+(j-1)*M) srec(k1,k2+M/2+(j-1)*M) srec(k1,k2+3*M/4+(j-1)*M)];
            G=G+s'*s;
        end
        re=4*[Rs(k1,k2); Rs(k1,k2+M/4); Rs(k1,k2+M/2);  Rs(k1,k2+3*M/4)]; %right hand side
        %solve 4x4 system
        mm=linsolve(G,re);
        m0(k1,k2)=mm(1);
        m0(k1,k2+M/4)=mm(2);
        m0(k1,k2+M/2)=mm(3);
        m0(k1,k2+3*M/4)=mm(4);
    end
end
elseif input_proj==4  % each second row and each third column acquired
                   % define G matrix (which decomposes into N*M/6 (6x6)-matrices) and solve the system to compute m
m0=zeros(N,M); %condition=zeros(N/2,M/2);
for k1=1:N/2
    for k2=1:M/3
        %construct G matrix of size 6 x 6
        G=zeros(6,6); re=zeros(6,1); %initialize G matrix and right-hand side
        for j=1:Nc
            s=[srec(k1+N/2,k2+(j-1)*M) srec(k1+N/2,k2+M/3+(j-1)*M) srec(k1+N/2,k2+2*M/3+(j-1)*M) srec(k1,k2+(j-1)*M)  srec(k1,k2+M/3+(j-1)*M) srec(k1,k2+2*M/3+(j-1)*M)];
            G=G+s'*s;
        end
        re=6*[Rs(k1+N/2,k2); Rs(k1+N/2,k2+M/3);Rs(k1+N/2,k2+2*M/3);  Rs(k1,k2);  Rs(k1,k2+M/3); Rs(k1,k2+2*M/3)]; %right hand side
        %solve 6x6 system
        %mm=linsolve(G,re);
        mm=pinv(G)*re;
        m0(k1+N/2,k2)=mm(1);
        m0(k1+N/2,k2+M/3)=mm(2);
        m0(k1+N/2,k2+2*M/3)=mm(3);
        m0(k1,k2)=mm(4);
        m0(k1,k2+M/3)=mm(5);
        m0(k1,k2+2*M/3)=mm(6);
    end
end
m1=m0; %reconstructed image
end

%renormalization of m not needed
mrec=m0;


% Improvement by one iteration step  of Algorithmus 4.3
for k=1:1
Yrec=zeros(N,M*Nc);
for j=1:Nc
        cY=fftshift(fft2(fftshift(m0.*srec(:,1+(j-1)*M:j*M))))./sqrt(M*N); 
        Yrec(:,1+(j-1)*M:j*M)=(ones(N,M)-P).*cY+ YP(:,1+(j-1)*M:j*M);
end

%improvement of m using all given and reconstructed measurements
mrec1=zeros(N,M); 
for j=1:Nc
    Yz=sqrt(M*N)*(ifftshift(ifft2(ifftshift(Yrec(:,1+(j-1)*M:j*M)))));
    mrec1=mrec1+conj(Yz).*Yz;
end
mrec=sqrt(mrec1);
end


end

