function [mrec,crec,Yrec] = mocca_iterative_mri(YP,P,Nc,M1,L,beta,iter)
% function Mocca algorithm (Algorithm 4.1 in Mocca paper) and SoS condition
%  for reconstruction from incomplete k-space data
% reconstruction of MRI images from incomplete k-space data (Algorithm 4.1)
%
%
%  Input: YP acquired coil image data YP=(Y^1.*P ...Y^Nc.*P ) of size
%          Nx(M*Nc), NxM dimension of the image to reconstruct (N, M even)
%         P projection NxM matrix, contains ones if corresponding Y^j entry is
%         acquired, zeros otherwise
%         Nc = number of coils
%         L= 2l+1 (odd), the square [-l,l]^2 of size L^2 is the support of
%         trigonometric polynomials for coil sensitivities
%         M1 = calibration matrix dimension in calibration area providing a
%         M1+L-1 box (M1 even)
%         N x M dimension of the image to reconstruct N even! M1 even!
%         iter: upper bound for number of iterations in Algorithm 4.3 to solve the least
%         squarse problem
%
% Output: mrec reconstructed image of size NxM,
%         crec matrix of coil coefficients  for all coils of size Lx(L*Nc)
%         Yrec =(Yrec^1, ... , Yrec^Nc) of size Nx(M*Nc) reconstructed k-space
%         data

%step 1: Build ESPIRIT matrix
[N,MM]=size(YP);
M=MM/Nc;
%fix the calibration data
Y1=[]; %extract the Nc centered calibration matrices of size M1+L-1 x M1+L-1
K1=floor(N/2)-floor((M1+L-1)/2); K2=floor(M/2)-floor((M1+L-1)/2);
for j=1:Nc
    Y1=[Y1 YP(K1+1:K1+M1+L-1,(j-1)*M+K2+1:(j-1)*M+K2+M1+L-1)];
end

%build partial ESPIRIT matrix from shifts of calibration data
l=floor((L-1)/2);
A1=[];
for j=1:Nc
Yb=Y1(:,1+(j-1)*(L+M1-1):j*(L+M1-1));
    for l1=-l:l
        for l2=-l:l
            Ybc=circshift(Yb,[l2,l1]);
            Ybcpart=Ybc(l+1:M1+l,l+1:M1+l); %partial calibration matrix blocks
            Ybv=reshape(Ybcpart,M1^2,1);   %calibration matrix blocks of size M1^2 x L^2
            A1=[A1 Ybv];
        end
    end
end

%Build the Mocca matrix of size M1^2 * Nc x L^2 * Nc
AMocca=[];
ysum=zeros(M1^2,L^2);
for j=1:Nc
    ysum=ysum+A1(:,1+(j-1)*L^2:j*L^2);
end
for k=1:Nc
    Z=[];
    for j=1:Nc
        if j==k
            Z=[Z A1(:,1+(k-1)*L^2:k*L^2)-ysum];
        else
        Z=[Z A1(:,1+(k-1)*L^2:k*L^2)];
        end
    end
    AMocca = [AMocca;Z];
end

% Compute svd of Mocca-Matrix to find the eigenvector c corresponding to
% smallest singular value
[U,S,V]=svd(AMocca);
cvec=V(:,end);                        %vector of sensitivity coefficients

% extract the sensitivity coefficients and reconstruct the sensitivity
% matrices
crec=[]; srec=[]; crec0=[];
for j=1:Nc
    cneu=cvec(1+(j-1)*L^2:j*L^2,1); %partion into coil sensitivity vectors
    cneu=reshape(cneu,L,L);
    crec=[crec cneu];

    % zero padding to obtain an NxM matrix
    K1=floor(N/2)-l; K2=floor(M/2)-l;
    zb=zeros(K1,L); zc=zeros(N+1,K2);
    cneu=[zb; cneu ; zb];
    cneu=[zc cneu zc];
    crec0=[crec0 cneu(1:N,1:M)];
    
    %reconstruct matrix of the sensitivity matrices s^j
    fc= sqrt(N*M)*ifftshift(ifft2(ifftshift(crec0(1:N,1+(j-1)*M:j*M))));
    srec=[srec sqrt(M*N)*fc];
end

% Compute vector d in step 4 of Algorithm 3.1 (step 1 of Algorithm 4.1)
d=zeros(N,M);
for j=1:Nc
    d=d+conj(srec(:,1+(j-1)*M:j*M)).*srec(:,1+(j-1)*M:j*M);
end

% compute d^+ (componentwise inverse of d if possible)
if min(d(:))> 10^(-12)
   dinv=ones(N,M)./d;
else
d=reshape(d,N*M,1);
 index=find(abs(d)<10^(-12));  % pay attention if d =0 for some components
 indexg=find(abs(d)>=10^(-12));
 dinv=ones(N*M,1); d0=zeros(N*M,1);
 dinv(indexg)=dinv(indexg)./d(indexg);
 dinv(index)=d0(index);
 dinv=reshape(dinv,N,M);
end


%step 2 of Algorithm 4.1: normalize sensitivities
d12=sqrt(dinv);
for j=1:Nc
    srec(:,1+(j-1)*M:j*M)= d12.*srec(:,1+(j-1)*M:j*M);
end

%step 3 of Algorithm 4.1: solve the least squares system
%compute m iteratively with Algorithm 4.3

%iteration initialization
m0=zeros(N,M); m1=m0;  Yrec=zeros(N,M*Nc);

% initial approximation of m from incomplete k-space data YP
for j=1:Nc
    cY=sqrt(M*N).*ifftshift(ifft2(ifftshift(YP(:,1+(j-1)*M:j*M))));
    m1=m1+ conj(srec(:,1+(j-1)*M:j*M)).*cY;
end

%iteration step (Yrec and m are updated at each step)
ind=0;
while and(abs(m0(:)-m1(:))>10^(-5),ind<iter)
    ind=ind+1;
    m2=zeros(M,N);
    for j=1:Nc
        cY=(fftshift(fft2(fftshift(m1.*srec(:,1+(j-1)*M:j*M)))))./sqrt(M*N);
        Yrec(:,1+(j-1)*M:j*M)=(1-beta).*(ones(M,N)-P).*cY+ YP(:,1+(j-1)*M:j*M);
        m2= m2+ sqrt(M*N).*conj(srec(:,1+(j-1)*M:j*M)).*ifftshift(ifft2(ifftshift(Yrec(:,1+(j-1)*M:j*M))));
    end
    m0=m1; m1=m2;
end
ind=ind

%renormalization of m not needed because we use sos condition
% mrec=m1.*d12;
mrec=m1;

end
