% test of Algorithm 3.1 in the paper 'MOCCA: A Fast Algorithm for Parallel MRI
% Reconstruction Using Model Based Coil Calibration for reconstruction of MRI images from
% incomplete k-space data' (by G. Plonka and Y. Riebe)
% here employing artificial data (randomly fixed sensitivities) and the image are
% reconstructed from complete k-space data


% Output: -original image
%         -reconstruced image mrec computed from the MOCCA algorithm 3.1
%         -pointwise error compared to original image


clc
clear all
close all

%Choice of parameters
L=5;            % LxL support size of the sensitivity in k-space
Nc=8;           % number of coils
M1=20;          % fixes size of calibration area: (M1+l-1)x(M1+L-1) box around


%load image
m = imread('pepper.png');        %magnetization matrix (example 1)
%m= imread('lena.jpg');           %magnetization matrix (example 2)
figure,imagesc(m); axis image
set(gca,'FontSize',20);
colormap(gray);
m=double(m);
m=m/255;
[N,M] = size(m);

%Simulate Nc coil sensitivity matrices (using bivariate trig. polynomials)
l=floor((L-1)/2);
c=zeros(N,Nc*M); s=[]; checkY=[]; Y=[]; %initialization of c and s, Y
% fix c_j randomly (complex)
for j=1:Nc
    for k1=-l:l
        for k2=-l:l
            %set random coefficients for cj
            c(floor((N+2)/2)+k1,floor((M+2)/2)+k2+(j-1)*M)=1*rand + 1i*rand; 
        end
    end
    %compute matrix of the sensitivity matrices 
    fc= sqrt(N*M).*ifftshift(ifft2(ifftshift(c(1:N,1+(j-1)*M:j*M)))); 
    s=[s fc]; 
    %compute the coil images in space domain
    checkY=[checkY m.*s(1:N,1+(j-1)*M:j*M)];  
    %compute the coil measurement matrices
    Y=[Y fftshift(fft2(fftshift(checkY(1:N,1+(j-1)*M:j*M))))]; 
end
Y=Y./sqrt(N*M);

[mrec,srec] = mocca_art(Y,M1,L,Nc); % apply MOCCA algorithm

% Comparison with reconstruction from complete data
 m1=m./max(m(:));            % normalization on [0,1]
 mrec=abs(mrec);
 mrec=mrec./max(mrec(:));    % normalization on [0,1]
 p = psnr1(m1,mrec)          % PSNR value of the error
 
 %Error map
 error=abs(mrec-m1)/max(m1(:));  %relative error
 figure('name','error map'), imagesc(error); axis off;
 colormap default
 colorbar
 set(gca,'FontSize',20);
 %caxis([0 0.2]);
 
%present the reconstruced image mrec
figure('name','reconstructed image with Mocca from complete data'),
 imagesc(mrec); axis off;
 set(gca,'FontSize',20);
 colormap(gray);
 set(gca,'LooseInset',get(gca,'TightInset'));
 saveas(gcf,'reconstructed_image.png')
 

function [mrec,crec] = mocca_art(Y,M1,L,Nc)
% reconstruction from complete artificial data (Algorithm 3.1)
%      
%
%  Input: Y coil image data Y=(Y^1 ...Y^Nc ) of size N times M*Nc
%         Nc = number of coils
%         L= 2l+1 = order of polynomial for coil sensitivities (each direction)
%         M1 = calibration matrix dimension in calibration area as M1+L-1 box
%         N x M dimension of the image to reconstruct N even! M1 even!
%         
%         
% Output: reconstructed image mrec, coil coefficients crec

%step 1: compute the partial blocks for the Mocca matrix 
[N,MM]=size(Y);
M=MM/Nc
% use only calibration data
Y1=[]; %extract Nc calibration matrices of size M1+L-1 x M1+L-1
K1=floor(N/2)-floor((M1+L-1)/2); K2=floor(M/2)-floor((M1+L-1)/2);
for j=1:Nc
    Y1=[Y1 Y(K1+1:K1+M1+L-1,(j-1)*M+K2+1:(j-1)*M+K2+M1+L-1)];
end
size(Y1)
l=floor((L-1)/2);
A1=[]; %build partial blocks of the Mocca matrix from shifts of calibration data
for j=1:Nc
    Yb=Y1(1:L+M1-1,1+(j-1)*(L+M1-1):j*(L+M1-1)); %partial matrices of size L+M1-1 x L+M1-1
    for l1=-l:l
        for l2=-l:l
            Ybc=circshift(Yb,[l2,l1]);
            Ybcpart=Ybc(l+1:M1+l,l+1:M1+l); %partial calibration matrix blocks
            Ybv=reshape(Ybcpart,M1^2,1);   %calibration matrix blocks of size M1^2 x L^2
            A1=[A1 Ybv];
        end
    end
end

%Build the Mocca matrix of size M1^2 * Nc x L^2 * Nc
AMocca=[];
ysum=zeros(M1^2,L^2);
for j=1:Nc
    ysum=ysum+A1(:,1+(j-1)*L^2:j*L^2);
end
for k=1:Nc
    Z=[];
    for j=1:Nc
        if j==k 
            Z=[Z A1(:,1+(k-1)*L^2:k*L^2)-ysum];
        else
        Z=[Z A1(:,1+(k-1)*L^2:k*L^2)];
        end
    end
    AMocca = [AMocca;Z];
end
% Compute svd of Mocca-Matrix to find the eigenvector c corresponding to
% smallest singualar value
[U,S,V]=svd(AMocca);
cvec=V(:,end); %vector of sensitivity coefficients


% extract the sensitivity coefficients and reconstruct the sensitivities
crec=zeros(N,Nc*M); srec=[]; 
for j=1:Nc
    cneu=cvec(1+(j-1)*L^2:j*L^2,1); %partition into coil sensitivity vectors
    cneu=reshape(cneu,L,L); 
    for k1=-l:l
        for k2=-l:l
        crec(floor((N+2)/2)+k1,floor((M+2)/2)+k2+(j-1)*M)=cneu(l+1+k1,l+1+k2); 
        end
    end
    %reconstruct matrix of the sensitivity matrices s^j
    fc= sqrt(N*M)*ifftshift(ifft2(ifftshift(crec(1:N,1+(j-1)*M:j*M)))); 
    srec=[srec fc]; 
end


% Compute  vector d in step 4 of Algorithm 3.1
d=zeros(N,M);
for j=1:Nc 
    d=d+conj(srec(:,1+(j-1)*M:j*M)).*srec(:,1+(j-1)*M:j*M);
end
% compute the inverse d^+
d=reshape(d,N*M,1);
 index=find(abs(d)<10^(-12))
 indexg=find(abs(d)>=10^(-12));
 dinv=ones(N*M,1); d0=zeros(N*M,1);
 dinv(indexg)=dinv(indexg)./d(indexg);
 dinv(index)=d0(index);
 dinv=reshape(dinv,N,M);
 
% compute m in step 5 of Algorithm 3.1
Yrec=[];
for j=1:Nc
Yrec=[Yrec ifftshift(ifft2(ifftshift(Y(1:N,1+(j-1)*M:j*M))))]; %inverse FFT to compute y(j)
end
Yrec=sqrt(N*M).*Yrec;
mrec=zeros(N,M);
for j=1:Nc
    mrec=mrec+conj(srec(:,1+(j-1)*M:j*M)).*Yrec(:,1+(j-1)*M:j*M);
end
mrec=dinv.*mrec;

end
    

