% test of Algorithm 4.1 in the paper 'MOCCA: A Fast Algorithm for Parallel MRI
% Reconstruction Using Model Based Coil Calibration for reconstruction of MRI images from
% incomplete k-space data' (by G. Plonka and Y. Riebe)
% here employing randomly chosen sensitivity coefficients, employing the
% iterative Algorithm 4.3 

% Output: -original image mrec0
%         -reconstruced image mrec from incomplete k-space data
%         -pointwise error compared to mrec0


clc
clear all
close all

%Choice of parameters
L=5;            % LxL support size of the sensitivity in k-space
Nc=8;           % number of coils
M1=20;          % fixes size of calibration area: (M1+l-1)x(M1+L-1) box around center
beta=0.00;      % regularization parameter in iteration of Algorithm 4.3
iter=500;       % number of maximal iterations in Algorithm 4.3
input_proj=1;   % Choice of projection to incomplete data (0=complete data, 1=each second column, 
                % 2=each second row and each second column, 3= each fourth column, 4=each
                % second row and each third column acquired)


%load original image
m = imread('pepper.png');       %magnetization matrix (example 1)
%m= imread('lena.jpg');           %magnetization matrix (example 2)
m=double(m);
m=m/255;
% represent the image in the first figure
figure('name','Original image'); imagesc(m); axis image
set(gca,'FontSize',20);
colormap(gray);

[N,M] = size(m);

%simulate Nc coil sensitivity matrices (using bivariate trig. polynomials)
l=floor((L-1)/2);
%initialization of c and s, Y
c=zeros(N,Nc*M); s=[]; checkY=[]; Y=[]; 
% fix c_j randomly (complex)
for j=1:Nc
    for k1=-l:l
        for k2=-l:l
            %set random coefficients for cj
            c(floor((N+2)/2)+k1,floor((M+2)/2)+k2+(j-1)*M)=1*(rand + 1i*rand); 
        end
    end
    %compute matrix the sensitivity matrices 
    fc= sqrt(N*M).*ifftshift(ifft2(ifftshift(c(1:N,1+(j-1)*M:j*M)))); 
    s=[s fc]; 
    %compute the coil images in space domain
    checkY=[checkY m.*s(1:N,1+(j-1)*M:j*M)];   
    %compute the coil measurement matrices
    Y=[Y fftshift(fft2(fftshift(checkY(1:N,1+(j-1)*M:j*M))))];   
end
Y=Y./sqrt(N*M); %normalization after FFT

% choice of projection matrix (unaquired data)

%input projection matrix (zero for each unacquired entry)
P=zeros(N,M);

if input_proj==0              %data completely acquired
    P=ones(N,M);
end
if input_proj==1              %each second column acquired
    P(:,1:2:end)=1;
 
elseif input_proj==2          %each second row and each second column acquired
    P(1:2:end,1:2:end)=1;
    
elseif input_proj==3          %each fourth column acquired
    P(:,1:4:end)=1;
    
elseif input_proj==4          %each second row and each third column acquired
    P(1:2:end,1:3:end)=1;
end

%calibration area always acquired    
K1=floor(N/2)-floor((M1+L-1)/2);
K2=floor(M/2)-floor((M1+L-1)/2);
P(K1+1:K1+L+M1-1,K2+1:K2+L+M1-1) = ones(L+M1-1,L+M1-1);

% Determine given k-space data
YP = [];
for j=1:Nc
YP = [YP Y(:,1+(j-1)*M:j*M).*P];  
end                             % YP only contains the acquired k-space data


% Mocca reconstruction according to Algorithm 4.1 in Mocca paper
[mrec,crec] = mocca_art_incomplete(YP,P,M1,L,Nc,beta,iter);

% Comparison with reconstruction from complete data
 m1=m./max(m(:));            % normalization on [0,1]
 mrec=abs(mrec);
 mrec=mrec./max(mrec(:));    % normalization on [0,1]
 p = psnr1(m1,mrec)          % PSNR value of the error
 
 
% Representation of the reconstructed image 
figure('name','Reconstructed image with Mocca using iterative algorithm');
imagesc(abs(mrec)); axis image
set(gca,'FontSize',20);
colormap(gray);

%Error map
 error=abs(mrec-m1/max(m1(:)));  %relative error
 figure('name','error map'), imagesc(error); axis off;
 colormap default
 colorbar
 set(gca,'FontSize',20);
 %saveas(gcf,'reconstruction_error_iterative.png')
 %caxis([0 0.2]);



function [mrec,crec] = mocca_art_incomplete(YP,P,M1,L,Nc,beta,iter)
% reconstruction from incomplete artificial data (Algorithm 4.1) using the
% iterative Algorithm 4.3
%      
%
%  Input: YP acquired coil image data YP=(Y^1.*P ...Y^Nc.*P ) of size
%          Nx(M*Nc), NxM dimension of the image to reconstruct (N, M even)
%         Nc = number of coils
%         L= 2l+1 = order of polynomial for coil sensitivities (each direction)
%         M1 = calibration matrix dimension in calibration area as M1+L-1 box
%         N x M dimension of the image to reconstruct N even! M1 even!
%         iter= maximal number of iterations in Algorithm 4.3
%         
% Output: reconstructed image mrec, coil coefficients crec

%step 1: compute the partial block matrices
[N,MM]=size(YP);
M=MM/Nc
%calibration data
Y1=[];              %extract calibration matrices of size M1+L-1 x M1+L-1
K1=floor(N/2)-floor((M1+L-1)/2); K2=floor(M/2)-floor((M1+L-1)/2);
for j=1:Nc
    Y1=[Y1 YP(K1+1:K1+M1+L-1,(j-1)*M+K2+1:(j-1)*M+K2+M1+L-1)];
end

l=floor((L-1)/2);
A1=[]; %build partial block matrices from shifts of calibration data
for j=1:Nc
    Yb=Y1(1:L+M1-1,1+(j-1)*(L+M1-1):j*(L+M1-1)); %partial matrices of size L+M1-1 x L+M1-1
    for l1=-l:l
        for l2=-l:l
            Ybc=circshift(Yb,[l2,l1]);     % order important
            Ybcpart=Ybc(l+1:M1+l,l+1:M1+l); %partial calibration matrix blocks
            Ybv=reshape(Ybcpart,M1^2,1);   %calibration matrix blocks of size M1^2 x L^2
            A1=[A1 Ybv];
        end
    end
end
size(A1)
%Build the Mocca matrix of size M1^2 * Nc x L^2 * Nc
AMocca=[];
ysum=zeros(M1^2,L^2);
for j=1:Nc
    ysum=ysum+A1(:,1+(j-1)*L^2:j*L^2);
end
for k=1:Nc
    Z=[];
    for j=1:Nc
        if j==k 
            Z=[Z A1(:,1+(k-1)*L^2:k*L^2)-ysum];
        else
        Z=[Z A1(:,1+(k-1)*L^2:k*L^2)];
        end
    end
    AMocca = [AMocca;Z];
end
% Compute svd of Mocca-Matrix to find the eigenvector c corresponding to
% smallest singualar value
[U,S,V]=svd(AMocca);
cvec=V(:,end); %vector of sensitivity coefficients


% extract the sensitivity coefficients and reconstruct the sensitivity
% matrices
crec=zeros(N,Nc*M); srec=[]; 
for j=1:Nc
    cneu=cvec(1+(j-1)*L^2:j*L^2,1); %partion into coil sensitivity vectors
    cneu=reshape(cneu,L,L); 
    for k1=-l:l
        for k2=-l:l
        crec(floor((N+2)/2)+k1,floor((M+2)/2)+k2+(j-1)*M)=cneu(l+1+k1,l+1+k2); 
        end
    end
    %reconstruct matrix of the sensitivity matrices sj
    fc= sqrt(N*M)*ifftshift(ifft2(ifftshift(crec(1:N,1+(j-1)*M:j*M))));
    srec=[srec fc]; 
end
% Compute  vector d in step 4 of Algorithm 3.1
d=zeros(N,M);
for j=1:Nc 
    d=d+conj(srec(:,1+(j-1)*M:j*M)).*srec(:,1+(j-1)*M:j*M);
end
d=reshape(d,N*M,1);
 index=find(abs(d)<10^(-12))
 indexg=find(abs(d)>=10^(-12));
 dinv=ones(N*M,1); d0=zeros(N*M,1);
 dinv(indexg)=dinv(indexg)./d(indexg);
 dinv(index)=d0(index);
 dinv=reshape(dinv,N,M);

%normalise sensitivities as in Algorithm 4.1
d12=sqrt(dinv);

for j=1:Nc 
    srec(:,1+(j-1)*M:j*M)= d12.*srec(:,1+(j-1)*M:j*M);
end

% compute m iteratively with Algorithm 4.3
m0=zeros(N,M); m1=m0;  Yrec=zeros(N,M*Nc);
%iteration initialization
for j=1:Nc   
    cY=sqrt(M*N).*ifftshift(ifft2(ifftshift(YP(:,1+(j-1)*M:j*M))));
    m1=m1+ conj(srec(:,1+(j-1)*M:j*M)).*cY;
end

ind=0;
%iteration
while and(abs(m0(:)-m1(:))>10^(-20),ind<iter)
    ind=ind+1;
    m2=zeros(M,N);
    for j=1:Nc
        cY=fftshift(fft2(fftshift(m1.*srec(:,1+(j-1)*M:j*M))))./sqrt(M*N);
        Yrec(:,1+(j-1)*M:j*M)=(1-beta).*(ones(M,N)-P).*cY+ YP(:,1+(j-1)*M:j*M);
        m2= m2+ conj(srec(:,1+(j-1)*M:j*M)).*ifftshift(ifft2(ifftshift(Yrec(:,1+(j-1)*M:j*M)))).*sqrt(M*N);
    end
    m0=m1; m1=m2;
end
ind=ind

%renormalization of m
mrec=m1.*(d12);

end

