% test of Algorithm 4.1 in the paper 'MOCCA: A Fast Algorithm for Parallel MRI
% Reconstruction Using Model Based Coil Calibration for reconstruction of MRI images from
% incomplete k-space data' (by G. Plonka and Y. Riebe)
% here employing Algorithm 4.3 for iterative solution of
% the equation system in step 3 of Algorithm 4.1.

% Output: -8 coil images (inverse Fourier transform from complete coil k-space data)
%         -reconstructed image mrec0 from complete k-space data computed with SOS condition
%         -reconstruced image mrec from incomplete k-space data
%         -pointwise error compared to mrec0
clc
clear all
close all


%Choice of parameters
L=5;            % LxL support size of the sensitivity in k-space
Nc=8;           % number of coils
M1=20;          % fixes size of calibration area: (M1+l-1)x(M1+L-1) box around center
beta=0.00;      % regularization parameter in iteration of Algorithm 4.3
iter=110;       % number of maximal iterations in Algorithm 4.3
input_data=1;   % change to cleaned data (with added zeros at boundary) for input_data=1
input_proj=2;   % Choice of projection to incomplete data (0=complete data, 1=each second column, 
                % 2=each second row and each second column, 3= each fourth column, 4=each
                % second row and each third column acquired)


load brain_8ch.mat;
DATA = DATA/max(max(max(abs(ifftshift(ifft2(ifftshift(DATA))))))) + eps;
[sx,sy,Nc] = size(DATA);
Y=reshape(DATA,sx,sy*Nc);

[N,MM] = size(Y);
 M=MM/Nc;
 
% Choose data brain_8ch.mat directly or cleaned data (with added zeros at boundary)
if input_data==1               %data brain_8ch.mat changed (8 added zero rows/columns)
   az=8; N0=N+2*az; M0=M+2*az; % az number of zeros added at each boundary
   Ycheck=[];  zr=zeros(az,M); zc=zeros(N0,az);
   for j=1:Nc
     Yz=sqrt(M*N)*ifftshift(ifft2(ifftshift(Y(:,1+(j-1)*M:j*M))));
     if or(j==1,j==8) Yz(194:200,80:119)=zeros(7,40); %cleaning of coil image 1 and 8 (bottom) 
     end
     if j==4 Yz(1:6,80:119)=zeros(6,40);              %cleaning of coil image 4 (ceil)
     end
     Yz=[zr; Yz ; zr];  % more zeros above and below 
     Yz=[zc Yz zc];     % more zeros left and right 
     Ycheck=[Ycheck Yz];
   end
    N=N0; M=M0;  %new values N,M
    Y=[];
    for j=1:Nc
      Yz=fftshift(fft2(fftshift(Ycheck(:,1+(j-1)*M:j*M))))./sqrt(M*N);
      Y=[Y Yz];     %cleaned data Y 
end
else 
    Ycheck=[];
    for j=1:Nc
    Yz=sqrt(M*N)*ifftshift(ifft2(ifftshift(Y(:,1+(j-1)*M:j*M))));
    Ycheck=[Ycheck Yz];
    end
end

%computation of mrec0 from complete k-space data with SOS
mrec0=zeros(N,M);         % initialize mrec0
for j=1:Nc
    Yz=Ycheck(:,1+(j-1)*M:j*M);
    mrec0=mrec0+conj(Yz).*Yz;   
end
mrec0=sqrt(mrec0); 
    
% illustration of coil images
K= floor(M*N*Nc/150);
Ycheck1=reshape(abs(Ycheck),M*N*Nc,1);
Yz0s=sort(Ycheck1(:));
index=find(Ycheck1>Yz0s(M*N*Nc-K));
Ycheck1(index)=Yz0s(M*N*Nc-K);
Ycheck1=reshape(Ycheck1,N,M,Nc);
    figure('name','coil images from complete k-space data'); 
    imshow3(Ycheck1); axis off;
set(gca,'FontSize',20);
colormap((gray(256)));

 % figure of reconstructed image from complete k-space data (with SoS)
 K=floor(M*N/100);                   % for proper scaling of grey values in the figure
 mrec00=reshape(mrec0, N*M,1);       % set the largest 1 percent of values to the maximum
 mrec0s=sort(mrec00(:));             % of the other values
 index=find(mrec00>mrec0s(M*N-K));
 mrec00(index)=mrec0s(M*N-K);
 mrec00=reshape(mrec00,N,M);

figure('name','image reconstruction from full data with SOS'),
imagesc(mrec00); axis off;
set(gca,'FontSize',20);
colormap(gray);



%input projection matrix (zero for each unacquired entry)
P=zeros(N,M);

if input_proj==0              %data completely acquired
    P=ones(N,M);
end
if input_proj==1              %each second column acquired
    P(:,1:2:end)=1;
 
elseif input_proj==2          %each second row and each second column acquired
    P(1:2:end,1:2:end)=1;
    
elseif input_proj==3            %each fourth column acquired
    P(:,1:4:end)=1;
    
elseif input_proj==4          %each second row and each third column acquired
    P(1:2:end,1:3:end)=1;
end

P1=P;   % projection matrix with structured unaquired data
%calibration area always acquired    
K1=floor(N/2)-floor((M1+L-1)/2);
K2=floor(M/2)-floor((M1+L-1)/2);
P(K1+1:K1+L+M1-1,K2+1:K2+L+M1-1) = ones(L+M1-1,L+M1-1);

% Determine given k-space data
YP = [];
for j=1:Nc
YP = [YP Y(:,1+(j-1)*M:j*M).*P];  
end                             % YP only contains the acquired k-space data


% Mocca reconstruction according to Algorithm 4.1 in Mocca paper using iteration in
% Algorithm 4.3
 [mrec,crec,Yrec] = mocca_iterative_mri(YP,P,Nc,M1,L,beta,iter);

 % Figure of reconstructed image mrec (1% of largest values projected to max of the other
 % values)
 
 mrec=abs(mrec);
 mrec=mrec./max(mrec(:));    % normalization on [0,1]
 K=floor(M*N/100);
 mrec1=reshape(mrec, M*N,1);
 mrec1s=sort(mrec(:));
 index=find(mrec>mrec1s(M*N-K));
 mrec1(index)=mrec1s(M*N-K);
 mrec1=reshape(mrec1,N,M);
 
 figure('name','Reconstructed image with Mocca using iterative algorithm'),
 imagesc(mrec1); axis off;
 set(gca,'FontSize',20);
 colormap(gray);
 set(gca,'LooseInset',get(gca,'TightInset'));
 saveas(gcf,'reconstruction_iterative.png')
 
% Comparison with reconstruction from complete data
 m1=mrec0./max(mrec0(:));    % normalization on [0,1]
 p = psnr1(m1,mrec)          % PSNR value of the error
 
 %Error map
 error=abs(mrec-m1/max(m1(:)));  %relative error
 figure('name','error map'), imagesc(error); axis off;
 colormap default
 colorbar
 set(gca,'FontSize',20);
 saveas(gcf,'reconstruction_error_iterative.png')
 %caxis([0 0.2]);
 

 
 
